import sys, os

import argparse

import subprocess
from setuptools_cuda_cpp import find_cuda_home_path

import logging

if __name__ == "__main__":

    logger = logging.getLogger("pyUAMMD")

    parser = argparse.ArgumentParser(description="Get UAMMD information")

    group = parser.add_mutually_exclusive_group()

    group.add_argument("--includers", action="store_true",help="Print the includers of the \
                                                                UAMMD and UAMMD-structured library")

    group.add_argument("--flags", action="store_true",help="Print the flags used to compile \
                                                            the UAMMD and UAMMD-structured library")

    group.add_argument("--libraries", action="store_true",help="Print the libraries used to link \
                                                                the UAMMD and UAMMD-structured library")

    group.add_argument("--cuda", action="store_true",help="Print the CUDA path")

    #Add arch optional argument
    parser.add_argument("--arch", type=str, help="CUDA architecture to compile for")

    args = parser.parse_args()

    if args.arch and not args.flags:
        logger.error("The architecture can only be set when printing the flags")
        raise RuntimeError("Incompatible arguments")

    if args.includers:

        # Try to import read environment variables UAMMD_PATH and UAMMD_STRUCTURED_PATH
        try:

            UAMMD_PATH = os.environ['UAMMD_PATH']

        except:

            logger.error("Environment variables UAMMD_STRUCTURED_PATH")
            raise RuntimeError("Environment variables not found")

        try:

            UAMMD_STRUCTURED_PATH = os.environ['UAMMD_STRUCTURED_PATH']

        except:

            logger.error("Environment variables UAMMD_PATH")
            raise RuntimeError("Environment variables not found")

        try:

            CUDA_PATH = find_cuda_home_path()

        except:
            raise RuntimeError('Can not find CUDA_HOME path')

        if CUDA_PATH is None:
            raise RuntimeError('Can not find CUDA_HOME path')

        INCLUDES = ['-I'+str(CUDA_PATH)+'/include/',
                    '-I'+UAMMD_PATH+'/src/',
                    '-I'+UAMMD_PATH+'/src/third_party/',
                    '-I'+UAMMD_STRUCTURED_PATH+'/']

        print(' '.join(INCLUDES))

    elif args.flags:

        try:

            NVCC_PATH = os.path.join(find_cuda_home_path(), 'bin', 'nvcc')

            gencode = []
            if args.arch:
                gencode.append(['-gencode',f'=arch=compute_{args.arch},code=sm_{args.arch}'])
            else:
                # Get the current list gpu capturing nvcc output
                p = subprocess.Popen([NVCC_PATH, '--list-gpu-code'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                out, err = p.communicate()

                archs = out.decode('utf-8').split('\n')[0:-1]
                archs = [arch.split("_")[-1] for arch in archs]

                for arch in archs:
                    gencode += ['-gencode',f'arch=compute_{arch},code=sm_{arch}']
        except:
            raise RuntimeError('Could not run nvcc')

        FLAGS = ['--expt-relaxed-constexpr',
                 '--expt-extended-lambda',
                 '-std=c++14',
                 '-O3',
                 '-DUAMMD_EXTENSIONS',
                 '-DMAXLOGLEVEL=5',
                 '-Xcompiler=\"-O3 -march=native -fPIC\"',
                 '-ccbin=g++'] + gencode

        print(' '.join(FLAGS))

    elif args.libraries:

        LIBS = ['-lcufft',
                '-llapacke',
                '-lcublas',
                '-lblas',
                '-lcurand',
                '-lcusolver',
                '-lstdc++fs']

        print(' '.join(LIBS))

    elif args.cuda:

        try:

            CUDA_PATH = find_cuda_home_path()

        except:
            raise RuntimeError('Can not find CUDA_HOME path')

        print(os.path.join(CUDA_PATH, 'bin', 'nvcc'))

    else:
        parser.print_help()
        sys.exit(0)







from .system import appendSystem
from .glb import appendGlobal
from .structure import appendStructure
from .state import appendState
from .integrator import appendIntegrator
from .simulationStep import appendSimulationStep
from .forceField import appendForceField
from .patchyParticles import appendPatchyParticles




import copy

import logging

from deepdiff import DeepDiff

from ..utils.update.groups import updateComponentSetGroupsLists
from ..utils.update.groups import appendGroups

from ..utils.update.ids import updateComponentSetIds

def appendIntegrator(topologyKey,
                     structureKey,
                     integratorKey,
                     sim,sim2app,mode,idOffset,modeOffset,availGroupTypes):

    logger = logging.getLogger("pyUAMMD")

    #########################
    #Appending integrator
    #Integrator structure:
    #   integratorKey: {
    #       "groups": {...} #Groups list, optional
    #       "integrator1": {
    #           "type": ["Class1","SubClass1"],
    #           "parameters": {
    #               "param1": "value1",
    #               "param2": "value2",
    #               ...
    #           }
    #       },
    #       "integrator2": {
    #           "type": ["Class2","SubClass2"],
    #           "parameters": {
    #               "param1": "value1",
    #               "param2": "value2",
    #               ...
    #           }
    #       },
    #       ...,
    #       "schedule": {
    #           "type": ["Schedule",integratorKey],
    #           "labels": ["order",integratorKey,"step"],
    #           "data": [[...,...,...],
    #                    [...,...,...],
    #                    ...]
    #       }
    #   }

    # Four cases:
    # 1) integrator NOT in sim and integrator NOT in sim2app
    # 2) integrator NOT in sim and integrator in sim2app
    # 3) integrator in sim and integrator NOT in sim2app
    # 4) integrator in sim and integrator in sim2app

    integratorInSim     = (integratorKey in sim)
    integratorInSim2app = (integratorKey in sim2app)

    structureInSim     = (topologyKey in sim)
    if structureInSim:
        structureInSim     = (structureKey in sim[topologyKey])

    structureInSim2app = (topologyKey in sim2app)
    if structureInSim2app:
        structureInSim2app    = (structureKey in sim2app[topologyKey])

    if (integratorInSim or integratorInSim2app) and not structureInSim:
        logger.warning("Structure must be present in at least one simulation")
        logger.error("Cannot merge integrator without any structure")
        raise Exception("Cannot merge integrator without structure")

    # 1) integrator NOT in sim and integrator NOT in sim2app
    if not integratorInSim and not integratorInSim2app:
        #Do nothing
        pass
    # 2) integrator NOT in sim and integrator in sim2app
    elif not integratorInSim and integratorInSim2app:
        #In this case the only thing to do is update the sim2app[integratorKey] groups list entries (if any)
        sim[integratorKey] = updateComponentSetGroupsLists(mode,sim2app[integratorKey],idOffset,modeOffset,availGroupTypes)
    # 3) integrator in sim and integrator NOT in sim2app
    elif integratorInSim and not integratorInSim2app:
        #Do nothing
        pass
    # 4) integrator in sim and integrator in sim2app
    elif integratorInSim and integratorInSim2app:
        #We append (after updating) the groupsLists in sim2app[integratorKey] to sim[integratorKey]
        appendGroups(mode,sim[integratorKey],sim2app[integratorKey],idOffset,modeOffset,availGroupTypes)

        integratorsSim     = [k for k in sim[integratorKey].keys() if "Schedule" not in sim[integratorKey][k]["type"]]
        integratorsSim2app = [k for k in sim2app[integratorKey].keys() if "Schedule" not in sim2app[integratorKey][k]["type"]]

        for inte2app in integratorsSim2app:
            if inte2app not in integratorsSim:
                sim[integratorKey][inte2app] = copy.deepcopy(sim2app[integratorKey][inte2app])
            else:
                #If are equal do nothing. Else raise error
                equal = not DeepDiff(sim[integratorKey][inte2app],
                                     sim2app[integratorKey][inte2app],
                                     ignore_order=True,report_repetition=True)
                if not equal:
                    logger.error("Cannot merge integrator with different parameters")
                    raise Exception("Cannot merge integrator with different parameters")

        #Merge schedule
        scheduleName = [k for k in sim[integratorKey].keys() if "Schedule" in sim[integratorKey][k]["type"]]
        if len(scheduleName) == 0:
            logger.error("Cannot merge integrator without schedule")
            raise Exception("Cannot merge integrator without schedule")
        elif len(scheduleName) > 1:
            logger.error("Cannot merge integrator with more than one schedule")
            raise Exception("Cannot merge integrator with more than one schedule")

        scheduleName = scheduleName[0]

        scheduleOffset = 0
        for inte in sim[integratorKey][scheduleName]["data"]:
            if inte[0] > scheduleOffset: #it assumes that the schedule is ordered and well formed
                scheduleOffset = inte[0]

        scheduleName2app = [k for k in sim2app[integratorKey].keys() if "Schedule" in sim2app[integratorKey][k]["type"]]
        if len(scheduleName2app) == 0:
            logger.error("Cannot merge integrator without schedule")
            raise Exception("Cannot merge integrator without schedule")
        elif len(scheduleName2app) > 1:
            logger.error("Cannot merge integrator with more than one schedule")
            raise Exception("Cannot merge integrator with more than one schedule")

        scheduleName2app = scheduleName2app[0]

        for inte in sim2app[integratorKey][scheduleName2app]["data"]:
            inteName = inte[1]
            if inteName not in integratorsSim:
                inte[0] += scheduleOffset
                sim[integratorKey][scheduleName]["data"].append(inte)

    #Integrator appended
    #########################



import copy

from deepdiff import DeepDiff

import logging

from ..utils.update.groups import updateComponentSetGroupsLists
from ..utils.update.groups import appendGroups

from ..utils.update.ids import updateComponentSetIds

def appendForceField(topologyKey,
                     structureKey,
                     forceFieldKey,
                     sim,sim2app,mode,idOffset,modeOffset,availGroupTypes,
                     id_labels,id_list_labels,type_labels,ignoredEntriesType=[]):

    logger = logging.getLogger("pyUAMMD")

    #########################
    #Appending force field
    #Force field structure:
    #   forceFieldKey: {
    #       "groups": {...} #Groups list, optional
    #       "interaction1": {
    #           "type": ["Class1","SubClass1"],
    #           "parameters": {
    #               "param1": "value",
    #               "param2": "value2",
    #               ...
    #           }
    #           "labels": ["label1","label2",...],
    #           "data": [[...],
    #                    [...],
    #                    ...]
    #       },
    #       "interaction2": {
    #           "type": ["Class2","SubClass2"],
    #           "parameters": {
    #               "param1": "value",
    #               "param2": "value2",
    #               ...
    #           }
    #           "labels": ["label1","label2",...],
    #           "data": [[...],
    #                    [...],
    #                    ...]
    #       },
    #       ...,
    #   }

    # Four cases:
    # 1) forceField NOT in sim and forceField NOT in sim2app
    # 2) forceField NOT in sim and forceField in sim2app
    # 3) forceField in sim and forceField NOT in sim2app
    # 4) forceField in sim and forceField in sim2app

    forceFieldInSim     = (topologyKey in sim)
    forceFieldInSim2app = (topologyKey in sim2app)

    if forceFieldInSim:
        forceFieldInSim     = (forceFieldKey in sim[topologyKey])
    if forceFieldInSim2app:
        forceFieldInSim2app = (forceFieldKey in sim2app[topologyKey])

    structureInSim     = (topologyKey in sim)
    if structureInSim:
        structureInSim     = (structureKey in sim[topologyKey])

    structureInSim2app = (topologyKey in sim2app)
    if structureInSim2app:
        structureInSim2app    = (structureKey in sim2app[topologyKey])

    if (forceFieldInSim or forceFieldInSim2app) and not (structureInSim or structureInSim2app):
        logger.warning("Structure must be present in at least one simulation")
        logger.error("Cannot merge forceField without any structure")
        raise Exception("Cannot merge forceField without structure")

    #Move ignored entries to ignoredEntries
    ignoredEntries = {}
    ignoredEntriesNames = []
    if forceFieldInSim2app:
        for entry in sim2app[topologyKey][forceFieldKey]:
            entryType = sim2app[topologyKey][forceFieldKey][entry]["type"][0]
            if entryType in ignoredEntriesType:
                ignoredEntriesNames.append(entry)

        for entry in ignoredEntriesNames:
            logger.debug("Moving ignored entry %s to ignoredEntries",entry)
            #Remove from sim2app (pop) and store in ignoredEntries
            ignoredEntries[entry] = sim2app[topologyKey][forceFieldKey].pop(entry)

    # 1) forceField NOT in sim and forceField NOT in sim2app
    if not forceFieldInSim and not forceFieldInSim2app:
        #Do nothing
        pass
    # 2) forceField NOT in sim and forceField in sim2app
    elif not forceFieldInSim and forceFieldInSim2app:
        #Check if the sim2app have particles defined.
        if structureInSim2app:
            #In this case, we first update the groups in sim2app and then de ids in the remainder entries
            sim[topologyKey][forceFieldKey] = updateComponentSetIds(mode,
                                                                  updateComponentSetGroupsLists(mode,
                                                                                                sim2app[topologyKey][forceFieldKey],
                                                                                                idOffset,modeOffset,availGroupTypes),
                                                                  idOffset,id_labels,id_list_labels)
        else:
            sim[topologyKey][forceFieldKey] = copy.deepcopy(sim2app[topologyKey][forceFieldKey])

    # 3) forceField in sim and forceField NOT in sim2app
    elif forceFieldInSim and not forceFieldInSim2app:
        #Do nothing
        pass
    # 4) forceField in sim and forceField in sim2app
    elif forceFieldInSim and forceFieldInSim2app:

        #Check if the sim2app have particles defined.
        if structureInSim2app:
            #We append (after updating) the groupsLists in sim2app[forceFieldKey] to sim[forceFieldKey]
            updatedForceField2app = updateComponentSetIds(mode,sim2app[topologyKey][forceFieldKey],idOffset,id_labels,id_list_labels)
            appendGroups(mode,sim[topologyKey][forceFieldKey],updatedForceField2app,idOffset,modeOffset,availGroupTypes)
        else:
            updatedForceField2app = copy.deepcopy(sim2app[topologyKey][forceFieldKey])
            appendGroups(mode,sim[topologyKey][forceFieldKey],updatedForceField2app,0,0,availGroupTypes)

        forceFieldSim     = [k for k in sim[topologyKey][forceFieldKey].keys() if sim[topologyKey][forceFieldKey][k]["type"] != ["Groups","GroupsList"]]
        forceFieldSim2app = [k for k in sim2app[topologyKey][forceFieldKey].keys() if sim2app[topologyKey][forceFieldKey][k]["type"] != ["Groups","GroupsList"]]

        for ff2app in forceFieldSim2app:
            if ff2app not in forceFieldSim:
                sim[topologyKey][forceFieldKey][ff2app] = copy.deepcopy(updatedForceField2app[ff2app])
            else:
                equalParam = not DeepDiff(sim[topologyKey][forceFieldKey][ff2app],
                                          updatedForceField2app[ff2app],
                                          ignore_order=True,report_repetition=True,
                                          exclude_paths=["root['data']","root['labels']"]) #Note data is excluded from comparison
                if equalParam:

                    if "data" not in sim[topologyKey][forceFieldKey][ff2app] or "data" not in updatedForceField2app[ff2app]:
                        continue

                    dataSim     = sim[topologyKey][forceFieldKey][ff2app]["data"]
                    dataSim2app = updatedForceField2app[ff2app]["data"]

                    if len(dataSim) != len(dataSim2app):
                        equalData = False
                    else:
                        equalData = not DeepDiff(dataSim,dataSim2app,ignore_order=True,report_repetition=True)

                    if equalData:
                        pass
                    else:

                        isSimIdDependent     = False
                        isSim2appIdDependent = False
                        for label in sim[topologyKey][forceFieldKey][ff2app]["labels"]:
                            if label in id_labels or label in id_list_labels:
                                isSimIdDependent = True
                                break
                        for label in updatedForceField2app[ff2app]["labels"]:
                            if label in id_labels or label in id_list_labels:
                                isSim2appIdDependent = True
                                break

                        isSimTypeDependent     = False
                        isSim2appTypeDependent = False
                        for label in sim[topologyKey][forceFieldKey][ff2app]["labels"]:
                            if label in type_labels:
                                isSimTypeDependent = True
                                break
                        for label in updatedForceField2app[ff2app]["labels"]:
                            if label in type_labels:
                                isSim2appTypeDependent = True
                                break

                        if isSimIdDependent != isSim2appIdDependent:
                            logger.error("Force field entry \"{}\" is id dependent in one of the simulations but not in the other".format(ff2app))
                            raise Exception("ForceField entry is id dependent in one of the simulations but not in the other")
                        else:
                            isIdDependent = isSimIdDependent

                        if isSimTypeDependent != isSim2appTypeDependent:
                            logger.error("Force field entry \"{}\" is type dependent in one of the simulations but not in the other".format(ff2app))
                            raise Exception("ForceField entry is type dependent in one of the simulations but not in the other")
                        else:
                            isTypeDependent = isSimTypeDependent

                        if isIdDependent or not isTypeDependent or (mode == "modelId"):
                            sim[topologyKey][forceFieldKey][ff2app]["data"].extend(updatedForceField2app[ff2app]["data"])
                        else:
                            if(mode == "simulationId"):

                                isSimulationIdPresentSim     = ("simulationId" in sim[topologyKey][forceFieldKey][ff2app]["labels"])
                                isSimulationIdPresentSim2app = ("simulationId" in updatedForceField2app[ff2app]["labels"])

                                if not isSimulationIdPresentSim and not isSimulationIdPresentSim2app:
                                    sim[topologyKey][forceFieldKey][ff2app]["labels"].append("simulationId")
                                    updatedForceField2app[ff2app]["labels"].append("simulationId")

                                    for i in range(len(sim[topologyKey][forceFieldKey][ff2app]["data"])):
                                        sim[topologyKey][forceFieldKey][ff2app]["data"][i].append(0)
                                    for i in range(len(updatedForceField2app[ff2app]["data"])):
                                        updatedForceField2app[ff2app]["data"][i].append(1)

                                    sim[topologyKey][forceFieldKey][ff2app]["data"].extend(updatedForceField2app[ff2app]["data"])

                                elif isSimulationIdPresentSim and not isSimulationIdPresentSim2app:
                                    updatedForceField2app[ff2app]["labels"].append("simulationId")

                                    for i in range(len(updatedForceField2app[ff2app]["data"])):
                                        updatedForceField2app[ff2app]["data"][i].append(modeOffset)

                                    sim[topologyKey][forceFieldKey][ff2app]["data"].extend(updatedForceField2app[ff2app]["data"])

                                elif not isSimulationIdPresentSim and isSimulationIdPresentSim2app:
                                    logger.error("SimulationId not present in reference simulation, but present in simulation to append.\
                                                 Force field entry: \"{}\"".format(ff2app))
                                    raise Exception("SimulationId not present in reference simulation, but present in simulation to append")
                                else:
                                    #Both are present
                                    logger.error("SimulationId present in both simulations. Force field entry: \"{}\"".format(ff2app))
                                    raise Exception("SimulationId present in both simulations")


                            else:
                                logger.error(f"This should not happen, please report it to the developers. \
                                             isIdDependent: {isIdDependent}, isTypeDependent: {isTypeDependent}, mode: {mode}")
                                raise Exception("This should not happen, please report it to the developers")

                else:
                    logger.error("Only force field entries which differ in data can be appended. Force field entry: \"{}\"".format(ff2app))
                    raise Exception("Only force field entries which differ in data can be appended")

    #If ignoreEntries is not empty, move back entries to sim2app
    if len(ignoredEntriesNames) > 0:
        for entry in ignoredEntriesNames:
            logger.debug("Moving back entry {} to simulation to append".format(entry))
            sim2app[topologyKey][forceFieldKey][entry] = ignoredEntries.pop(entry)

    #Force field appended
    #########################



import copy

from deepdiff import DeepDiff

import logging

from ..utils.update.groups import appendGroups

from ..utils.update.ids import updateComponentSetIds

def appendSimulationStep(topologyKey,
                         structureKey,
                         simulationStepKey,
                         sim,sim2app,mode,idOffset,modeOffset,availGroupTypes,
                         id_labels,id_list_labels):

    logger = logging.getLogger("pyUAMMD")

    #########################
    #Appending simulation step
    #Simulation steps structure:
    #   simulationStepKey: {
    #       "groups": {...} #Groups list, optional
    #       "step1": {
    #           "type": ["Class1","SubClass1"],
    #           "parameters": {
    #               "intervalStep": "value",
    #               "startStep": "value", #optional
    #               "endStep": "value", #optional
    #               "param2": "value2",
    #               ...
    #           }
    #           "labels": ["label1","label2",...],
    #           "data": [[...],
    #                    [...],
    #                    ...]
    #       },
    #       "step2": {
    #           "type": ["Class2","SubClass2"],
    #           "parameters": {
    #               "intervalStep": "value",
    #               "startStep": "value", #optional
    #               "endStep": "value", #optional
    #               "param2": "value2",
    #               ...
    #           }
    #           "labels": ["label1","label2",...],
    #           "data": [[...],
    #                    [...],
    #                    ...]
    #       },
    #       ...,
    #   }

    # Four cases:
    # 1) simulationStep NOT in sim and simulationStep NOT in sim2app
    # 2) simulationStep NOT in sim and simulationStep in sim2app
    # 3) simulationStep in sim and simulationStep NOT in sim2app
    # 4) simulationStep in sim and simulationStep in sim2app

    simulationStepInSim     = (simulationStepKey in sim)
    simulationStepInSim2app = (simulationStepKey in sim2app)

    structureInSim     = (topologyKey in sim)
    if structureInSim:
        structureInSim     = (structureKey in sim[topologyKey])

    structureInSim2app = (topologyKey in sim2app)
    if structureInSim2app:
        structureInSim2app    = (structureKey in sim2app[topologyKey])

    if (simulationStepInSim or simulationStepInSim2app) and not structureInSim:
        logger.warning("Structure must be present in at least one simulation")
        logger.error("Cannot merge simulationStep without any structure")
        raise Exception("Cannot merge simulationStep without structure")

    # 1) simulationStep NOT in sim and simulationStep NOT in sim2app
    if not simulationStepInSim and not simulationStepInSim2app:
        #Do nothing
        pass
    # 2) simulationStep NOT in sim and simulationStep in sim2app
    elif not simulationStepInSim and simulationStepInSim2app:
        #Check if the sim2app have particles defined.
        if structureInSim2app:
            #In this case, we first update the groups in sim2app and then de ids in the remainder entries
            sim[simulationStepKey] = updateComponentSetIds(mode,
                                                               updateComponentSetGroupsLists(mode,
                                                                                             sim2app[simulationStepKey],
                                                                                             idOffset,modeOffset,availGroupTypes),
                                                               idOffset,id_labels,id_list_labels)
        else:
            sim[simulationStepKey] = copy.deepcopy(sim2app[simulationStepKey])

    # 3) simulationStep in sim and simulationStep NOT in sim2app
    elif simulationStepInSim and not simulationStepInSim2app:
        #Do nothing
        pass
    # 4) simulationStep in sim and simulationStep in sim2app
    elif simulationStepInSim and simulationStepInSim2app:

        #Check if the sim2app have particles defined.
        if structureInSim2app:
            #We append (after updating) the groupsLists in sim2app[simulationStepKey] to sim[simulationStepKey]
            updatedSimulationStep2app = updateComponentSetIds(mode,sim2app[simulationStepKey],idOffset,id_labels,id_list_labels)
            appendGroups(mode,sim[simulationStepKey],updatedSimulationStep2app,idOffset,modeOffset,availGroupTypes)
        else:
            updatedSimulationStep2app = copy.deepcopy(sim2app[simulationStepKey])
            appendGroups(mode,sim[simulationStepKey],updatedSimulationStep2app,0,0,availGroupTypes)

        simulationStepsSim     = [k for k in sim[simulationStepKey].keys() if sim[simulationStepKey][k]["type"] != ["Groups","GroupsList"]]
        simulationStepsSim2app = [k for k in sim2app[simulationStepKey].keys() if sim2app[simulationStepKey][k]["type"] != ["Groups","GroupsList"]]

        for simStep2app in simulationStepsSim2app:
            if simStep2app not in simulationStepsSim:
                sim[simulationStepKey][simStep2app] = copy.deepcopy(updatedSimulationStep2app[simStep2app])
            else:
                equalParam = not DeepDiff(sim[simulationStepKey][simStep2app],
                                          updatedSimulationStep2app[simStep2app],
                                          ignore_order=True,report_repetition=True,
                                          exclude_paths=["root['data']"]) #Note data is excluded from comparison
                if equalParam:

                    if "data" not in sim[simulationStepKey][simStep2app] and "data" not in updatedSimulationStep2app[simStep2app]:
                        continue

                    dataSim     = sim[simulationStepKey][simStep2app]["data"]
                    dataSim2app = updatedSimulationStep2app[simStep2app]["data"]

                    if len(dataSim) != len(dataSim2app):
                        equalData = False
                    else:
                        equalData = not DeepDiff(dataSim,dataSim2app,ignore_order=True,report_repetition=True)

                    if equalData:
                        pass
                    else:
                        sim[simulationStepKey][simStep2app]["data"].extend(updatedSimulationStep2app[simStep2app]["data"])

                else:
                    logger.error("Only simulation step entries which differ in data can be appended. Simulation step entry: \"{}\"".format(ff2app))
                    raise Exception("Only simulation step entries which differ in data can be appended")

    #Simulation step appended
    #########################



import copy

import logging

from ..utils.common import getDataEntryLabelIndex

def appendStructure(topologyKey,
                    structureKey,
                    sim,sim2app,mode):

    logger = logging.getLogger("pyUAMMD")

    idOffset   = 0
    modeOffset = 0

    #########################
    #Appending structure
    #Structure structure:
    #   structureKey: {
    #       "labels": ["id","type","resId","chainId","modelId","simulationId"]
    #       "data": [[...],
    #                [...],
    #                ...]

    # Four cases:
    # 1) structure NOT in sim and structure NOT in sim2app
    # 2) structure NOT in sim and structure in sim2app
    # 3) structure in sim and structure NOT in sim2app
    # 4) structure in sim and structure in sim2app

    structureInSim     = (topologyKey in sim)
    structureInSim2app = (topologyKey in sim2app)

    if structureInSim:
        structureInSim     = (structureKey in sim[topologyKey])
    if structureInSim2app:
        structureInSim2app = (structureKey in sim2app[topologyKey])

    #Create offsets if reference has structure
    if structureInSim:

        #Computing id offset
        idIndex = getDataEntryLabelIndex(sim[topologyKey][structureKey],"id")

        idOffset = 0
        for d in sim[topologyKey][structureKey]["data"]:
            if d[idIndex] > idOffset:
                idOffset = d[idIndex]
        idOffset += 1

        modeOffset = 0
        if mode == "modelId":
            if "modelId" not in sim[topologyKey][structureKey]["labels"]:
                sim[topologyKey][structureKey]["labels"].append("modelId")
                for d in sim[topologyKey][structureKey]["data"]:
                    d.append(0)

            modeIndex = getDataEntryLabelIndex(sim[topologyKey][structureKey],"modelId")
            for d in sim[topologyKey][structureKey]["data"]:
                if d[modeIndex] > modeOffset:
                    modeOffset = d[modeIndex]
            modeOffset += 1
        elif mode == "simulationId":
            if "simulationId" not in sim[topologyKey][structureKey]["labels"]:
                sim[topologyKey][structureKey]["labels"].append("simulationId")
                for d in sim[topologyKey][structureKey]["data"]:
                    d.append(0)

            modeIndex = getDataEntryLabelIndex(sim[topologyKey][structureKey],"simulationId")
            for d in sim[topologyKey][structureKey]["data"]:
                if d[modeIndex] > modeOffset:
                    modeOffset = d[modeIndex]
            modeOffset += 1
        else:
            logger.error(f"Appending mode {mode} not implemented")
            raise Exception("Appending mode not implemented")

    # 1) structure NOT in sim and structure NOT in sim2app
    if not structureInSim and not structureInSim2app:
        #Do nothing
        pass
    # 2) structure NOT in sim and structure in sim2app
    elif not structureInSim and structureInSim2app:
        if topologyKey not in sim:
            sim[topologyKey] = {}
        sim[topologyKey][structureKey] = copy.deepcopy(sim2app[topologyKey][structureKey])

        idOffset   = 0
        modeOffset = 0
    # 3) structure in sim and structure NOT in sim2app
    elif structureInSim and not structureInSim2app:
        #Do nothing
        pass
    # 4) structure in sim and structure in sim2app
    elif structureInSim and structureInSim2app:

        #Handle labels
        simLabels     = sim[topologyKey][structureKey]["labels"]
        sim2appLabels = sim2app[topologyKey][structureKey]["labels"]

        availLabels = ["id","type","resId","chainId","modelId","simulationId"]

        for label in simLabels:
            if label not in availLabels:
                logger.error(f"Label {label} not available for structure.")
                raise Exception("Label not available for structure.")

        for label in sim2appLabels:
            if label not in availLabels:
                logger.error(f"Label {label} not available for structure.")
                raise Exception("Label not available for structure.")

        for label in availLabels:
            labelInSim     = (label in simLabels)
            labelInSim2app = (label in sim2appLabels)

            if not labelInSim and not labelInSim2app:
                #Do nothing
                pass
            elif not labelInSim and labelInSim2app:
                sim[topologyKey][structureKey]["labels"].append(label)
                for d in sim[topologyKey][structureKey]["data"]:
                    d.append(0)
            elif labelInSim and not labelInSim2app:
                sim2app[topologyKey][structureKey]["labels"].append(label)
                for d in sim2app[topologyKey][structureKey]["data"]:
                    d.append(0)
            elif labelInSim and labelInSim2app:
                #Do nothing
                pass

        #Data merging
        for structInfo in sim2app[topologyKey][structureKey]["data"]:
            d = []
            for label in sim[topologyKey][structureKey]["labels"]:
                if label == "id":
                    d.append(structInfo[sim2appLabels.index(label)] + idOffset)
                elif label == mode:
                    d.append(structInfo[sim2appLabels.index(label)] + modeOffset)
                else:
                    d.append(structInfo[sim2appLabels.index(label)])
            sim[topologyKey][structureKey]["data"].append(d)

    #Structure appended
    #########################

    return idOffset,modeOffset

